module DenseList exposing
    ( DenseList
    , decoder
    , empty
    , encodeReversed
    , foldLeftUntilErrorDecoder
    , insert
    )

import Bidict exposing (Bidict)
import Json.Decode as Decode exposing (Decoder)
import Json.Encode as Encode


type DenseList value
    = DenseList
        { nextKey : Int
        , values : Bidict Int value
        , keys : List Int
        }


empty : DenseList value
empty =
    DenseList
        { nextKey = 0
        , values = Bidict.empty
        , keys = []
        }


insert : comparable -> DenseList comparable -> DenseList comparable
insert value (DenseList model) =
    case Bidict.getKeys value model.values of
        key :: _ ->
            DenseList
                { model | keys = key :: model.keys }

        [] ->
            DenseList
                { model
                    | nextKey = model.nextKey + 1
                    , values = Bidict.insert model.nextKey value model.values
                    , keys = model.nextKey :: model.keys
                }


encodeReversed : (value -> Encode.Value) -> DenseList value -> Encode.Value
encodeReversed encodeValue (DenseList model) =
    Encode.object
        [ ( "nextKey", Encode.int model.nextKey )
        , ( "values", Bidict.encode Encode.int encodeValue model.values )
        , ( "keys", Encode.list Encode.int (List.reverse model.keys) )
        ]


decoder : Decoder comparable -> Decoder (DenseList comparable)
decoder valueDecoder =
    Decode.map3
        (\nextKey values keys ->
            DenseList { nextKey = nextKey, values = values, keys = keys }
        )
        (Decode.field "nextKey" Decode.int)
        (Decode.field "values" (Bidict.decoder Decode.int valueDecoder))
        (Decode.field "keys" (Decode.list Decode.int))


foldLeftUntilErrorDecoder :
    Decoder comparable
    -> (comparable -> Result error a)
    -> (a -> b -> b)
    -> b
    -> Decoder ( DenseList comparable, b )
foldLeftUntilErrorDecoder valueDecoder resultOp foldOp foldInit =
    Decode.map3
        (\nextKey values keys ->
            ( DenseList
                { nextKey = nextKey
                , values = values
                , keys = keys
                }
            , foldLeftUntilError resultOp foldOp values keys foldInit
            )
        )
        (Decode.field "nextKey" Decode.int)
        (Decode.field "values" (Bidict.decoder Decode.int valueDecoder))
        (Decode.field "keys" (Decode.list Decode.int))


foldLeftUntilError : (comparable -> Result error a) -> (a -> b -> b) -> Bidict Int comparable -> List Int -> b -> b
foldLeftUntilError resultOp foldOp values keys output =
    case keys of
        key :: moreKeys ->
            case Bidict.getValue key values of
                Just value ->
                    case resultOp value of
                        Ok foldValue ->
                            foldLeftUntilError
                                resultOp
                                foldOp
                                values
                                moreKeys
                                (foldOp foldValue output)

                        Err _ ->
                            output

                Nothing ->
                    output

        [] ->
            output
